# Latent class regression

data <- read.table("/Users/Thierry/Documents/R/Data/Compstat/program_effectiveness.txt",
                   header=TRUE)
attach(data)

w<- GPA
u<-PSI
v<-TUCE

#-------------------------------------
# Q1

logistic <- function(x){
	A<-exp(x)
	return(A/(1+A))
}

# Observed-data log-likelihood
loglik<- function(theta,w,u,v,lambda=0){
  # theta= vector of parameters
  # Lambda= regularization hyperparameter
	beta1=theta[1:2]
	beta0=theta[3:4]
	alpha=theta[5:6]
	sigma1=theta[7]
	sigma0=theta[8]
	n<- length(w)
	logL<- -lambda*t(alpha)%*%alpha
	for(i in 1:n){
		P <- logistic(alpha[1]+alpha[2]*v[i])
		phi1<-dnorm(w[i],mean=beta1[1]+beta1[2]*u[i],sd=abs(sigma1))
		phi0<-dnorm(w[i],mean=beta0[1]+beta0[2]*u[i],sd=abs(sigma0))
		g<-P*phi1+(1-P)*phi0
		logL=logL+log(g)
	}
	return(logL)
}

reg0<-lm(w~u)
beta<-reg0$coefficients
sigma<-sqrt(mean(reg0$residuals^2))
theta0<-c(beta,beta,c(0,0),sigma,sigma)

N<-100 # number or runs
f_best<--Inf
for(i in 1:N){
  theta<-theta0+0.1*rnorm(8)
  opt<-optim(par=theta,fn=loglik,method = "BFGS",
             control=list(fnscale=-1,trace=0),w=w,u=u,v=v,lambda=0.2)
  print(c(i,opt$value,f_best))
  if(opt$value>f_best){
    opt_best<-opt
    f_best<-opt$value
  }
}
theta<-opt_best$par
beta1=theta[1:2]
beta0=theta[3:4]
alpha=theta[5:6]
sigma1=theta[7]
sigma0=theta[8]
P<-logistic(alpha[1]+alpha[2]*v)


#-------------------------
# Q2


# Part of the Q function that depends on alpha1 and alpha2
F_alpha <- function(alpha,z,v,lambda){
	x<-alpha[1]+alpha[2]*v
	res<- z*x - log(1+exp(x)) - lambda*t(alpha)%*%alpha
	return(sum(res))
}

em_latent_class_reg <- function(w,u,v,theta,
                                lambda,epsi,trace){

n<-length(w)
# initialization
logL0<-loglik(theta,w,u,v,lambda)
beta1=theta[1:2]
beta0=theta[3:4]
alpha=theta[5:6]
sigma1=theta[7]
sigma0=theta[8]
go_on<-TRUE
t<-0
z<-vector(mode="numeric",length=n)

while(go_on){
	t<-t+1
	
# E-step

for(i in 1:n){
	phi1<-dnorm(w[i],mean=beta1[1]+beta1[2]*u[i],sd=abs(sigma1))
	phi0<-dnorm(w[i],mean=beta0[1]+beta0[2]*u[i],sd=abs(sigma0))
	P <- logistic(alpha[1]+alpha[2]*v[i])
	z[i]<-phi1*P/(phi1*P + phi0*(1-P))
}

# M-step

reg1<-lm(w ~ u, weights=z)
beta1<-reg1$coefficients
S<-sum(z)
sigma1 <- sqrt(sum(z*reg1$residuals^2)/S)

reg0<-lm(w ~ u, weights=1-z)
beta0<-reg0$coefficients
sigma0 <- sqrt(sum((1-z)*reg0$residuals^2)/(n-S))

opt<- optim(par=alpha,fn=F_alpha,control=list(fnscale=-1),z=z,v=v,lambda=lambda)
alpha<-opt$par

theta<-c(beta1,beta0,alpha,sigma1,sigma0)
logL<-loglik(theta,w,u,v,lambda)

if (logL-logL0 < epsi){
			go_on <- FALSE
			}
		logL0 <- logL
		if(trace==1) print(c(t,logL))
}

return(list(loglik=logL,theta=theta,z=z))

}




epsi<-1e-6
trace<-0
lambda<-0.2

N<-100 # number or runs
f_best<--Inf
for(i in 1:N){
	theta<-theta0+0.1*rnorm(8)
	em<-em_latent_class_reg(w,u,v,theta,lambda,epsi,trace)
	print(c(i,em$loglik,f_best))
	if(em$loglik>f_best){
		em_best<-em
		f_best<-em$loglik
	}
}

z<-em_best$z
plot(u[z<0.5],w[z<0.5])
points(u[z>0.5],w[z>0.5],pch=2)





